(* listset.ml: a set of unique elements that is kept in a sorted
   list. *)

(* an interface module for the Make functor below *)
module type ELEM_SIG = sig
  type elem_t;;
  val compare : elem_t -> elem_t -> int;;
  val elem_string : elem_t -> string;;
end;;

(* Make functor to create sets of unique elements backed by a
   list. ElType is a parameter module which carries the element type,
   a comparison function and function to stringify elements. *)
module Make(ElType : ELEM_SIG) = struct

  type set =                            (* type for the list *)
    | Empty                             (* end of list *)
    | Node of {data : ElType.elem_t;    (* reference element type *)
               next : set;}
  ;;

  let empty = Empty;;                   (* empty element *)

  let rec add set elem =                (* add to set *)
    match set with
    | Empty -> Node{data=elem;
                    next=Empty}
    | Node(node) ->
       let diff =                       (* compute a difference *)
         ElType.compare elem node.data  (* using the compare function *)
       in
       if diff = 0 then                 (* redundant *)
         set
       else if diff < 0 then            (* found insertion point *)
         Node{data=elem; next=set}
       else
         let rest =                     (* move ahead *)
           add node.next elem
         in 
         Node{node with next=rest}
  ;;

  let rec to_string set =               (* create a string version of the set *)
    let buf = Buffer.create 256 in
    let rec build set =
      match set with
      | Empty -> ()
      | Node(node) ->
         let estr =
           ElType.elem_string node.data (* use stringify function in param module *)
         in
         Buffer.add_string buf estr;
         Buffer.add_string buf " ";
         build node.next;
    in
    build set;
    Buffer.contents buf
  ;;

end;;

